/* adal.c */

#include <Pilot.h>

#include <Graffiti.h>		/* for GrfSetState() */
#include <SysEvtMgr.h>		/* for EvtFlushNextPenStroke() */
#include <SysEvtPrv.h>		/* for EvtGetPenBtnList() */

#ifdef USE_SONYCHARS
#include <SonyChars.h>
#include <SonyHwrOEMIDs.h>
#define sysFtrNumOEMCompanyID 20
#endif

#ifdef USE_HSEXT
#include <HsExt.h>
#endif

#ifdef USE_TRGCHARS
#define vchrCFlashMin 0x1500
#include <TrgChars.h>
#endif

#ifdef USE_VGA
#include "MyVga.h"
#endif

#include "adal.h"
#include "adal_rcp.h"


/*
 * macros
 */

#if defined(__GNUC__) && defined(USE_A4_REGISTER)
#define ARG(arg)
#define ARG_(arg)
#define SAVE_A4_REGISTER void *old_g = g;
#define RESTORE_A4_REGISTER g = old_g;
#else
#define ARG(arg) arg
#define ARG_(arg) arg,
#define SAVE_A4_REGISTER
#define RESTORE_A4_REGISTER
#endif

#ifdef __GNUC__
register void *reg_a7 asm("%sp");
#define __getA7() (long) reg_a7
#define __setA7(a7) reg_a7 = (void *) (a7)
#else /* __MWERKS__ */
long __getA7(void) = { 0x200F };
void __setA7(long a7:__d0) = { 0x2E40 };
#endif

#define CENTER_X(r) ((r).topLeft.x + (r).extent.x / 2 - 2)
#define CENTER_Y(r) ((r).topLeft.y + (r).extent.y / 2)

#define printable(c) ((c) >= ' ' && (c) <= 'z')


/*
 * constants
 */

#define HACK_DB_TYPE 'HACK'

#define FTR_ORIG TRAP1_ID
#define FTR_GLOBALS 9000

#define APP_DB_TYPE 'appl'
#define PANEL_DB_TYPE 'panl'
#define PREFS_CREATOR 'pref'

#define CALC_CREATOR 'calc'
#define ADVANCED_CALC_CREATOR 'HsPr'
#define ADVANCED_CALC_TYPE 'PrAp'

#define DA_DB_TYPE 'DAcc'
#define DA_RSRC_TYPE 'code'
#define DA_RSRC_ID 1000

#define MIN_STACK_SIZE 256

#define KEYED_DELAY 30

#define LIST_HEIGHT 14
#define LIST_ITEM_HEIGHT 11
#define LIST_LEFT 2
#define LIST_BOTTOM 158

#define SAVE_WIDTH 80
#define SAVE_HEIGHT 160

#define WAIT_MESSAGE "preparing..."
#define WAIT_X 4
#define WAIT_Y 147

#ifdef USE_VGA
#define VGA_LIST_ITEM_HEIGHT 16
#define VGA_LIST_MARGIN 3
#define VGA_SAVE_WIDTH 120
#define VGA_WAIT_X 5
#define VGA_WAIT_Y_DELTA 19
#endif

#define NONE_STRING "none"
#define ELLIPSIS_STRING "..."
#define ELLIPSIS_LENGTH 3
#define FLASH_SUFFIX " *"
#define EXT_SUFFIX " *"

#define FLASHPRO_HEAP_ID 4
#define JACKFLASH_HEAP_ID 3

enum ret_code {
    DO_NOTHING,
    DO_LAUNCH,
    SWITCH_MODE
};


/*
 * types
 */

#define DB_INFO_LENGTH 4

struct db_info {
    UInt card_no;
    LocalID db_id;
    union {
	unsigned long l;
	char s[DB_INFO_LENGTH];
    } name;
};

struct globals {
    ULong creator;
    struct adal_prefs prefs;
    int tap_mode;
    int mode;
    int new_mode;
    FormPtr saved_form;
    Word saved_form_id;
    WinHandle saved_active_win;
    WinHandle saved_draw_win;
    FontID saved_font;
    FormPtr form;
    ListPtr list;
    struct db_info *db_list;
    int db_cnt;
    UInt card_no;
    LocalID db_id;
    char name[dmDBNameLength];
    char keyed_name[dmDBNameLength];
    int keyed_length;
    int keyed_first;
    int keyed_last;
    Boolean history;
    Boolean reversed;
#ifdef USE_VGA
    Boolean vga_1to1;
#endif
};

#if defined(__GNUC__) && defined(USE_A4_REGISTER)
register struct globals *g asm("%a4");
#endif


/*
 * functions
 */

static int find_part(PointType *start, PointType *end)
{
    PenBtnInfoPtr buttons;
    UInt num_buttons, i;
    RectanglePtr apps_rect;
    struct adal_prefs prefs;

    buttons = EvtGetPenBtnList(&num_buttons);
    apps_rect = NULL;
    for (i = 0; i < num_buttons; ++i) {
	if (buttons[i].asciiCode == launchChr) {
	    apps_rect = &buttons[i].boundsR;
	    break;
	}
    }

    if (apps_rect == NULL
     || ! RctPtInRectangle(start->x, start->y, apps_rect)) {
	return MODE_NONE;
    }

    get_prefs(&prefs);

    if (RctPtInRectangle(end->x, end->y, apps_rect)) {
	return (&prefs.button1)[
		(start->x >= CENTER_X(*apps_rect)) * 2
	      + (start->y >= CENTER_Y(*apps_rect))];
    } else {
	if (end->x < apps_rect->topLeft.x + apps_rect->extent.x) {
	    if (end->y < apps_rect->topLeft.y) {
		return prefs.drag1;
	    } else {
		return prefs.drag5;
	    }
	} else {
	    if (end->y < apps_rect->topLeft.y) {
		return prefs.drag2;
	    } else if (end->y < apps_rect->topLeft.y + apps_rect->extent.y) {
		return prefs.drag3;
	    } else {
		return prefs.drag4;
	    }
	}
    }

    return MODE_NONE;
}

static Boolean is_device_locked(void)
{
    DWord rom_version;
    SystemPreferencesTypeV10 sys_prefs;

    if (FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_version) != noErr) {
	return false;
    } else if (sysGetROMVerMajor(rom_version) < 2) {
	PrefGetPreferences((SystemPreferencesPtr) &sys_prefs);
	return sys_prefs.deviceLocked;
    } else {
	return PrefGetPreference(prefDeviceLocked);
    }
}

static char to_lower(char c)
{
    return (c >= 'A' && c <= 'Z') ? (c - 'A' + 'a') : c;
}

static Err get_icon_name(
	ARG_(struct globals *g) UInt card_no, LocalID db_id, char *name)
{
    Err err;
    UInt attr;
    ULong type;
    DmOpenRef db;
    VoidHand h;
    VoidPtr p;

    err = DmDatabaseInfo(card_no, db_id, name, &attr,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, &type, NULL);
    if (err == noErr
     && (attr & dmHdrAttrResDB) != 0
     && ((type == APP_DB_TYPE && g->prefs.app_name_lookup)
      || (type == DA_DB_TYPE && g->prefs.da_name_lookup))) {
	db = DmOpenDatabase(card_no, db_id, dmModeReadOnly);
	if (db != NULL) {
	    h = DmGet1Resource(ainRsc, ainID);
	    if (h != NULL) {
		p = MemHandleLock(h);
		if (p != NULL) {
		    StrNCopy(name, p, dmDBNameLength - 1);
		    name[dmDBNameLength - 1] = '\0';
		    MemHandleUnlock(h);
		}
		DmReleaseResource(h);
	    }
	    DmCloseDatabase(db);
	}
    }
    return err;
}

static Err get_db_name(ARG_(struct globals *g) int i)
{
    return get_icon_name(
	    ARG_(g) g->db_list[i].card_no, g->db_list[i].db_id, g->name);
}


/*
 * db_list management
 */

static int compare_names(
	ARG_(struct globals *g) struct db_info *db_info, char *name, int i)
{
    if ((g->prefs.sort_method & SORT_SLOW) == SORT_FAST) {
	if (db_info->name.l < g->db_list[i].name.l) {
	    return -1;
	}
	if (db_info->name.l > g->db_list[i].name.l) {
	    return 1;
	}
    }

    if (get_db_name(ARG_(g) i) != noErr) {
	return 0;
    }

    if ((g->prefs.sort_method & SORT_JAPANESE_FIRST) != 0) {
	if ((name[0] & 0x80) != 0 && (g->name[0] & 0x80) == 0) {
	    return -1;
	}
	if ((name[0] & 0x80) == 0 && (g->name[0] & 0x80) != 0) {
	    return 1;
	}
    }

    return StrCaselessCompare(name, g->name);
}

static void put_db_info(ARG_(struct globals *g) Byte sort_order)
{
    char *name = g->keyed_name;
    struct db_info db_info;
    int i, j, k, cmp;

    if (get_icon_name(ARG_(g) g->card_no, g->db_id, name) == noErr) {
	db_info.card_no = g->card_no;
	db_info.db_id = g->db_id;
	db_info.name.l = 0;
	for (j = 0; j < DB_INFO_LENGTH && name[j] != '\0'; ++j) {
	    db_info.name.s[j] = to_lower(name[j]);
	}
	if ((g->prefs.sort_method & SORT_JAPANESE_FIRST) != 0) {
	    db_info.name.s[0] ^= 0x80;
	}

	if (sort_order == HISTORY_ALPHABETICAL) {
	    i = 0;
	    j = g->db_cnt;
	    while (i < j) {
		k = (i + j) / 2;
		cmp = compare_names(ARG_(g) &db_info, name, k);
		if (cmp < 0) {
		    j = k;
		} else if (k > i) {
		    i = k;
		} else {
		    i = j;
		}
	    }
	} else if (sort_order == HISTORY_REVERSE_USED) {
	    i = 0;
	} else {
	    i = g->db_cnt;
	}

	for (j = g->db_cnt; j > i; --j) {
	    g->db_list[j] = g->db_list[j - 1];
	}
	g->db_list[i] = db_info;
	++g->db_cnt;
    }
}

static void gen_db_list(ARG(struct globals *g))
{
    int mode;
    ULong type;
    Boolean new_search;
    DmSearchStateType state;
    int cnt;

    g->db_list = NULL;
    g->db_cnt = 0;

    cnt = 0;
    for (mode = g->mode & MODE_BOTH, type = APP_DB_TYPE;
	 mode != MODE_NONE; mode >>= 1, type = DA_DB_TYPE) {
	if (mode & 1) {
	    for (new_search = true;
		 DmGetNextDatabaseByTypeCreator(new_search, &state,
			type, 0, false, &g->card_no, &g->db_id) == noErr;
		 new_search = false) {
		++cnt;
	    }
	}
    }

    if (cnt == 0
     || (g->db_list = MemPtrNew(sizeof(struct db_info) * cnt)) == NULL) {
	return;
    }

    for (mode = g->mode & MODE_BOTH, type = APP_DB_TYPE;
	 mode != MODE_NONE; mode >>= 1, type = DA_DB_TYPE) {
	if (mode & 1) {
	    for (new_search = true;
		 g->db_cnt < cnt
	      && DmGetNextDatabaseByTypeCreator(new_search, &state,
			type, 0, false, &g->card_no, &g->db_id) == noErr;
		 new_search = false) {
		put_db_info(ARG_(g) HISTORY_ALPHABETICAL);
	    }
	}
    }
}

static void put_history_entry(
	ARG_(struct globals *g) struct history_entry *entry, Byte sort_order)
{
    Boolean new_search;
    DmSearchStateType state;
    UInt card_no, card_nos[5];
    LocalID db_id, db_ids[5];
    char name[dmDBNameLength];
    int i, j;

    for (i = 0; i < 5; ++i) {
	db_ids[i] = NULL;
    }

    for (new_search = true;
	 DmGetNextDatabaseByTypeCreator(new_search, &state,
		entry->type, entry->creator, false, &card_no, &db_id) == noErr;
	 new_search = false) {
	i = 4;

	if (card_no == entry->card_no
	 && db_id == entry->db_id) {
	    i = 0;
	} else if (DmDatabaseInfo(card_no, db_id, name, NULL,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL) == noErr
	 && StrCompare(name, entry->name) == 0) {
	    i = 3;

	    if (card_no == entry->card_no) {
		if (MemLocalIDKind(db_id) == entry->db_id_kind
		 && db_ids[1] == NULL) {
		    i = 1;
		} else if (db_ids[2] == NULL) {
		    i = 2;
		}
	    }
	}

	if (db_ids[i] == NULL) {
	    card_nos[i] = card_no;
	    db_ids[i] = db_id;
	}
    }

    for (i = 0; i < 5; ++i) {
	if (db_ids[i] != NULL) {
	    for (j = 0; j < g->db_cnt; ++j) {
		if (g->db_list[j].card_no == card_nos[i]
		 && g->db_list[j].db_id == db_ids[i]) {
		    break;
		}
	    }
	    if (j == g->db_cnt) {
		g->card_no = card_nos[i];
		g->db_id = db_ids[i];
		put_db_info(ARG_(g) sort_order);
		return;
	    }
	}
    }
}

static void gen_history_list(
	ARG_(struct globals *g) int mode, int max, Byte sort_order)
{
    DmOpenRef db;
    Boolean res_db;
    VoidHand h;
    struct history_record *p;
    int cnt, i;

    g->db_list = NULL;
    g->db_cnt = 0;

    db = DmOpenDatabaseByTypeCreator(HISTORY_DB_TYPE, CREATOR, dmModeReadOnly);
    if (db != NULL) {
	if (DmOpenDatabaseInfo(db, NULL, NULL, NULL, NULL, &res_db) == noErr
	 && res_db != false
	 && (h = DmGet1Resource(HISTORY_RSRC_TYPE, HISTORY_RSRC_ID)) != NULL) {
	    if (MemHandleSize(h) == HISTORY_RSRC_SIZE
	     && (p = (struct history_record *) MemHandleLock(h)) != NULL) {
		if (p->entry_size == sizeof(struct history_entry)
		 && p->entry_cnt > 0 && p->entry_cnt <= HISTORY_MAX_COUNT) {
		    cnt = min(p->entry_cnt, max);
		    g->db_list = MemPtrNew(sizeof(struct db_info) * cnt);
		    if (g->db_list != NULL) {
			for (i = 0; i < p->entry_cnt && g->db_cnt < max; ++i) {
			    if (mode == MODE_HISTORY
			     || (mode == MODE_APP_HISTORY
			      && p->entries[i].type == APP_DB_TYPE)
			     || (mode == MODE_DA_HISTORY
			      && p->entries[i].type == DA_DB_TYPE)) {
				put_history_entry(
					ARG_(g) &p->entries[i], sort_order);
			    }
			}
		    }
		}
		MemHandleUnlock(h);
	    }
	    DmReleaseResource(h);
	}
	DmCloseDatabase(db);
    }
}

static void free_db_list(ARG(struct globals *g))
{
    if (g->db_list != NULL) {
	MemPtrFree(g->db_list);
    }
}


/*
 * user interaction
 */

static void draw_list_item(Word item_no, RectanglePtr bounds, CharPtr *dummy)
{
    struct globals ARG_(*g) *gg;
    char *name, *suffix;
    UInt heap_id;
    Int width, new_width, ellipsis_width, length;
    Boolean fits;

    SAVE_A4_REGISTER

    name = NONE_STRING;
    suffix = NULL;

    if (FtrGet(CREATOR, FTR_GLOBALS, (DWord *) &gg) == noErr
     && (g = gg) != NULL && g->creator == CREATOR
     && item_no >= 0 && item_no < g->db_cnt
     && get_db_name(ARG_(g) item_no) == noErr) {

	name = g->name;

	if (g->db_list[item_no].card_no == 0) {
	    if ((g->prefs.mark_flash_ext & MARK_FLASH) != 0
	     && MemLocalIDKind(g->db_list[item_no].db_id) == memIDPtr) {
		heap_id = MemPtrHeapID(MemLocalIDToPtr(
			g->db_list[item_no].db_id,
			g->db_list[item_no].card_no));
		if (heap_id == FLASHPRO_HEAP_ID
		 || heap_id == JACKFLASH_HEAP_ID) {
		    suffix = FLASH_SUFFIX;
		}
	    }
	} else {
	    if ((g->prefs.mark_flash_ext & MARK_EXT) != 0) {
		suffix = EXT_SUFFIX;
	    }
	}
    }

    width = bounds->extent.x;
    if (suffix != NULL) {
	width -= FntCharsWidth(suffix, StrLen(suffix));
    }

    new_width = width;
    length = StrLen(name);
    FntCharsInWidth(name, &new_width, &length, &fits);
    if (fits) {
	WinDrawChars(name, length, bounds->topLeft.x, bounds->topLeft.y);
    } else {
	ellipsis_width = FntCharsWidth(ELLIPSIS_STRING, ELLIPSIS_LENGTH);
	new_width = width - ellipsis_width;
	FntCharsInWidth(name, &new_width, &length, &fits);
	WinDrawChars(name, length, bounds->topLeft.x, bounds->topLeft.y);
	WinDrawChars(ELLIPSIS_STRING, ELLIPSIS_LENGTH,
		bounds->topLeft.x + new_width, bounds->topLeft.y);
	new_width += ellipsis_width;
    }

    if (suffix != NULL) {
	WinDrawChars(suffix, StrLen(suffix),
		bounds->topLeft.x + new_width, bounds->topLeft.y);
    }

    RESTORE_A4_REGISTER
}

static void clear_input(ARG_(struct globals *g) ListPtr list)
{
    g->keyed_length = 0;

    if (g->db_cnt == 0) {
	LstSetSelection(list, noListSelection);
    } else if (g->reversed) {
	LstSetSelection(list, g->db_cnt - 1);
    } else {
	LstSetSelection(list, 0);
    }
}

static int match_name(ARG_(struct globals *g) int i, int *jp)
{
    int j;
    char c1, c2;

    for (j = 0; j < g->keyed_length; ++j) {
	c1 = g->keyed_name[j];
	if (j < DB_INFO_LENGTH) {
	    c2 = g->db_list[i].name.s[j];
	    if (j == 0 && (g->prefs.sort_method & SORT_JAPANESE_FIRST) != 0) {
		c2 ^= 0x80;
	    }
	} else {
	    if (j == DB_INFO_LENGTH) {
		if (get_db_name(ARG_(g) i) != noErr) {
		    *jp = j;
		    return c1;
		}
	    }
	    c2 = to_lower(g->name[j]);
	}
	if (c1 != c2) {
	    *jp = j;
	    return c1 - c2;
	}
    }
    return 0;
}

static Boolean match_name2(ARG_(struct globals *g) int i, int j_max)
{
    int j, k;
    char c1, c2;

    if (get_db_name(ARG_(g) i) != noErr) {
	return false;
    }

    for (j = 0, k = 0; j < g->keyed_length; ++j, ++k) {
	c1 = g->keyed_name[j];
	c2 = to_lower(g->name[k]);
	if (j < j_max) {
	    if (c1 != c2) { return false; }
	} else {
	    while (c1 != c2) {
		if (c2 == '\0') { return false; }
		c2 = to_lower(g->name[++k]);
	    }
	}
    }
    return true;
}

static void find_db(ARG(struct globals *g))
{
    int first, last, i, j, j_max, j_first, j_last, match;

    first = -1;
    last = 0;

    j_max = 0;
    j_first = 0;
    j_last = -1;

    for (i = 0; i < g->db_cnt; ++i) {
	match = match_name(ARG_(g) i, &j);
	if (match == 0) {
	    if (first < 0) { first = i; }
	    last = i;
	} else if (j >= j_max) {
	    if (j > j_max) {
		j_max = j;
		j_first = i;
	    }
	    j_last = i;
	} else if (match < 0 && g->history) {
	    break;
	}
    }

    if (first < 0) {
	for (i = j_first; i <= j_last; ++i) {
	    if (match_name2(ARG_(g) i, j_max)) {
		if (first < 0) { first = i; }
		last = i;
	    }
	}
    }

    g->keyed_first = first;
    g->keyed_last = last;
}

static short event_loop(ARG(struct globals *g))
{
    EventType event;
    Word chr;
    ListPtr list = g->list;
    short new_sel, list_height = LIST_HEIGHT;

#ifdef USE_SONYCHARS
    DWord company;
    Boolean is_sony = false, jog_paged = false;

    if (FtrGet(sysFtrCreator, sysFtrNumOEMCompanyID, &company) == noErr
     && company == sonyHwrOEMCompanyID_Sony) {
	is_sony = true;
    }
#endif

#ifdef USE_VGA
    if (g->vga_1to1) {
	list_height = LstGetVisibleItems(list);
    }
#endif

    while (g->new_mode == MODE_NONE) {
	EvtGetEvent(&event, evtWaitForever);

	chr = (event.eType == keyDownEvent) ? event.data.keyDown.chr : 0;
	if (event.eType == appStopEvent
	 || chr == menuChr || chr == launchChr
	 || chr == findChr || ChrIsHardKey(chr)
#ifdef USE_HSEXT
	 || chr == hsChrCardStatusChg
	 || (chr >= hsChrCardUserFirst && chr <= hsChrCardUserLast)
#endif
#ifdef USE_TRGCHARS
	 || chr == vchrSilkToggle
#endif
	) {
	    EvtAddEventToQueue(&event);
	    break;
	} else if (chr == backspaceChr) {
	    if (g->prefs.backspace == BACKSPACE_PREVIOUS_ITEM) {
		new_sel = LstGetSelection(list) - 1;
		if (new_sel < 0) { new_sel = 0; }
		LstSetSelection(list, new_sel);
	    } else if (g->prefs.backspace == BACKSPACE_CLEAR_CHAR
		    && g->keyed_length >= 2) {
		--g->keyed_length;
		goto find;		/* !!! */
	    } else {
		clear_input(ARG_(g) list);
	    }
	} else if (chr == linefeedChr || chr == crChr) {
	    return LstGetSelection(list);
	} else if (chr == pageUpChr || chr == upArrowChr) {
	    new_sel = LstGetSelection(list) -
		((g->prefs.page_key == PAGEKEY_SCROLL_LINE) ? 1 : list_height);
	    if (new_sel < 0) { new_sel = 0; }
	    LstSetSelection(list, new_sel);
	} else if (chr == pageDownChr || chr == downArrowChr) {
	    new_sel = LstGetSelection(list) +
		((g->prefs.page_key == PAGEKEY_SCROLL_LINE) ? 1 : list_height);
	    if (new_sel >= g->db_cnt) { new_sel = g->db_cnt - 1; }
	    LstSetSelection(list, new_sel);
	} else if (chr == leftArrowChr || chr == prevFieldChr
#ifdef USE_SONYCHARS
		|| (is_sony && chr == vchrJogUp)
#endif
	) {
	    new_sel = LstGetSelection(list) - 1;
	    if (new_sel < 0) { new_sel = 0; }
	    LstSetSelection(list, new_sel);
	} else if ((chr == spaceChr && g->prefs.space == SPACE_NEXT_ITEM)
		|| chr == rightArrowChr || chr == nextFieldChr
#ifdef USE_SONYCHARS
		|| (is_sony && chr == vchrJogDown)
#endif
	) {
	    new_sel = LstGetSelection(list) + 1;
	    if (new_sel >= g->db_cnt) { new_sel = g->db_cnt - 1; }
	    LstSetSelection(list, new_sel);
#ifdef USE_SONYCHARS
	} else if (is_sony && chr == vchrJogPushedUp) {
	    jog_paged = true;
	    new_sel = LstGetSelection(list) - list_height;
	    if (new_sel < 0) { new_sel = 0; }
	    LstSetSelection(list, new_sel);
	} else if (is_sony && chr == vchrJogPushedDown) {
	    jog_paged = true;
	    new_sel = LstGetSelection(list) + list_height;
	    if (new_sel >= g->db_cnt) { new_sel = g->db_cnt - 1; }
	    LstSetSelection(list, new_sel);
	} else if (is_sony && chr == vchrJogPush) {
	    jog_paged = false;
	} else if (is_sony && chr == vchrJogRelease) {
	    if (jog_paged == false) {
		return LstGetSelection(list);
	    }
#endif
	} else if (chr == escapeChr
#ifdef USE_SONYCHARS
		|| (is_sony && chr == vchrJogBack)
#endif
	) {
	    break;
	} else if (printable(chr) && g->keyed_length < sizeof g->keyed_name) {
	    g->keyed_name[g->keyed_length++] = to_lower(chr);
	find:
	    find_db(ARG(g));
	    if (g->keyed_first >= 0) {
		if (g->prefs.input_feedback) {
		    LstSetSelection(list,
			    g->reversed ? g->keyed_first : g->keyed_last);
		    SysTaskDelay(KEYED_DELAY);
		}
		if (g->keyed_first == g->keyed_last) {
		    return g->keyed_first;
		} else {
		    LstSetSelection(list,
			    g->reversed ? g->keyed_last : g->keyed_first);
		}
	    }
	} else {
	    if (! SysHandleEvent(&event)
	     && ! FrmHandleEvent(g->form, &event)) {
		if (event.eType == lstSelectEvent) {
		    return event.data.lstSelect.selection;
		} else if (event.eType == penDownEvent) {
		    break;
		}
	    }
	}

	if (g->new_mode == g->tap_mode || g->new_mode == g->mode) {
	    g->tap_mode = g->new_mode;
	    if (g->prefs.tap_again == TAPAGAIN_ROTATE_APP_DA) {
		if (g->mode == MODE_HISTORY) {
		    g->new_mode = MODE_APP;
		} else if (g->mode == MODE_APP) {
		    g->new_mode = MODE_DA;
		} else {
		    g->new_mode = MODE_HISTORY;
		}
	    } else if (g->prefs.tap_again == TAPAGAIN_ROTATE_DA_APP) {
		if (g->mode == MODE_HISTORY) {
		    g->new_mode = MODE_DA;
		} else if (g->mode == MODE_DA) {
		    g->new_mode = MODE_APP;
		} else {
		    g->new_mode = MODE_HISTORY;
		}
	    } else if (g->prefs.tap_again == TAPAGAIN_ROTATE_BOTH) {
		if (g->mode == MODE_APP) {
		    g->new_mode = MODE_APP_HISTORY;
		} else if (g->mode == MODE_APP_HISTORY) {
		    g->new_mode = MODE_APP;
		} else if (g->mode == MODE_DA) {
		    g->new_mode = MODE_DA_HISTORY;
		} else if (g->mode == MODE_DA_HISTORY) {
		    g->new_mode = MODE_DA;
		} else if (g->mode == MODE_HISTORY) {
		    g->new_mode = MODE_BOTH;
		} else {
		    g->new_mode = MODE_HISTORY;
		}
	    }
	} else if (g->new_mode != MODE_NONE) {
	    g->tap_mode = g->new_mode;
	}

	if (g->new_mode == g->mode) {
	    if (g->prefs.tap_again == TAPAGAIN_DO_NOTHING) {
		g->new_mode = MODE_NONE;
	    } else if (g->prefs.tap_again == TAPAGAIN_CLEAR_INPUT) {
		clear_input(ARG_(g) list);
		g->new_mode = MODE_NONE;
	    } else if (g->prefs.tap_again == TAPAGAIN_OS_LAUNCHER) {
		event.eType = keyDownEvent;
		event.data.keyDown.chr = launchChr;
		event.data.keyDown.keyCode = 0;
		event.data.keyDown.modifiers = commandKeyMask;
		EvtAddEventToQueue(&event);
	    } else if (g->prefs.tap_again == TAPAGAIN_LAST_ITEM) {
		if (g->mode == MODE_APP || g->mode == MODE_APP_HISTORY) {
		    g->new_mode = MODE_LAST_APP;
		} else if (g->mode == MODE_DA || g->mode == MODE_DA_HISTORY) {
		    g->new_mode = MODE_LAST_DA;
		} else {
		    g->new_mode = MODE_NONE;
		}
	    }
	}
    }

    return noListSelection;
}

static int open_launcher(ARG(struct globals *g))
{
    DmOpenRef db;
    Boolean res_db;
    WinHandle save;
    RectangleType r;
    Err err;
    ListPtr list;
    int list_cnt, list_height, list_item_height, list_left, list_bottom;
    short sel;
    int ret_code = DO_NOTHING;

#ifdef USE_VGA
    DWord vga_version;
    VgaScreenModeType mode;
    VgaRotateModeType rotation;

    g->vga_1to1 = false;

    if (FtrGet(TRGSysFtrID, TRGVgaFtrNum, &vga_version) == noErr) {
	VgaGetScreenMode(&mode, &rotation);
	if (mode == screenMode1To1) {
	    g->vga_1to1 = true;
	}
    }
#endif

    db = DmOpenDatabaseByTypeCreator(HACK_DB_TYPE, CREATOR, dmModeReadOnly);
    if (db != NULL) {
	if (DmOpenDatabaseInfo(db, NULL, NULL, NULL, NULL, &res_db) == noErr
	 && res_db != false) {
	    g->form = FrmInitForm(MAIN_FORM_ID);
	    FrmSetActiveForm(g->form);

	    GrfSetState(false, false, false);

	    RctSetRectangle(&r, 0, 0, SAVE_WIDTH, SAVE_HEIGHT);
#ifdef USE_VGA
	    if (g->vga_1to1) {
		VgaFormModify(g->form, vgaFormModify160To240);
		WinGetDisplayExtent(&r.extent.x, &r.extent.y);
		WinSetWindowBounds(FrmGetWindowHandle(g->form), &r);
		r.extent.x = VGA_SAVE_WIDTH;
	    }
#endif
	    save = WinSaveBits(&r, &err);

#ifdef USE_VGA
	    if (g->vga_1to1) {
		FntSetFont(VgaBaseToVgaFont(stdFont));
		WinDrawChars(WAIT_MESSAGE, StrLen(WAIT_MESSAGE),
			VGA_WAIT_X, r.extent.y - VGA_WAIT_Y_DELTA);
	    } else
#endif
	    WinDrawChars(WAIT_MESSAGE, StrLen(WAIT_MESSAGE), WAIT_X, WAIT_Y);

	    g->history = false;
	    g->reversed = false;
	    if (g->mode == MODE_HISTORY
	     || g->mode == MODE_APP_HISTORY
	     || g->mode == MODE_DA_HISTORY) {
		g->history = true;
		if (g->prefs.history_order == HISTORY_REVERSE_USED) {
		    g->reversed = true;
		}
		gen_history_list(ARG_(g) g->mode,
			g->prefs.history_count, g->prefs.history_order);
	    } else {
		gen_db_list(ARG(g));
	    }

	    g->list = list = FrmGetObjectPtr(
		    g->form, FrmGetObjectIndex(g->form, MAIN_LIST_ID));
	    LstSetDrawFunction(list, &draw_list_item);
	    list_cnt = (g->db_cnt > 0) ? g->db_cnt : 1;
	    LstSetListChoices(list, NULL, list_cnt);
	    list_height = LIST_HEIGHT;
	    list_item_height = LIST_ITEM_HEIGHT;
	    list_left = LIST_LEFT;
	    list_bottom = LIST_BOTTOM;
#ifdef USE_VGA
	    if (g->vga_1to1) {
		list_height = (r.extent.y - VGA_LIST_MARGIN * 2)
			/ VGA_LIST_ITEM_HEIGHT;
		list_item_height = VGA_LIST_ITEM_HEIGHT;
		list_left = VGA_LIST_MARGIN;
		list_bottom = r.extent.y - VGA_LIST_MARGIN;
	    }
#endif
	    if (list_height > list_cnt) {
		list_height = list_cnt;
	    }
	    LstSetPosition(list,
		    list_left, list_bottom - list_item_height * list_height);
	    LstSetHeight(list, list_height);
	    clear_input(ARG_(g) list);
	    LstEraseList(list);
	    LstDrawList(list);

	    sel = event_loop(ARG(g));

	    if (sel >= 0 && sel < g->db_cnt) {
		g->card_no = g->db_list[sel].card_no;
		g->db_id = g->db_list[sel].db_id;
		ret_code = DO_LAUNCH;
	    }

	    if ((g->new_mode >= MODE_APP && g->new_mode <= MODE_DA_HISTORY)
	     && g->new_mode != g->mode) {
		ret_code = SWITCH_MODE;
	    }

	    free_db_list(ARG(g));

	    WinRestoreBits(save, 0, 0);

	    FrmDeleteForm(g->form);
	}
	DmCloseDatabase(db);
    }

    return ret_code;
}


/*
 * launching
 */

static void record_history(
	UInt card_no, LocalID db_id, ULong type, ULong creator, char *name)
{
    DmOpenRef db;
    Boolean res_db;
    VoidHand h, new_h;
    struct history_record *p;
    struct history_record record;
    struct history_entry entry;
    int count, i;

    entry.card_no = card_no;
    entry.db_id_kind = MemLocalIDKind(db_id);
    entry.db_id = db_id;
    entry.type = type;
    entry.creator = creator;
    MemSet(entry.name, sizeof entry.name, 0);
    StrCopy(entry.name, name);

    db = DmOpenDatabaseByTypeCreator(HISTORY_DB_TYPE, CREATOR, dmModeReadWrite);
    if (db == NULL) {
	if (DmCreateDatabase(HISTORY_CARD_NO, HISTORY_DB_NAME,
		CREATOR, HISTORY_DB_TYPE, true) == noErr) {
	    db = DmOpenDatabaseByTypeCreator(
		    HISTORY_DB_TYPE, CREATOR, dmModeReadWrite);
	}
    }
    if (db != NULL) {
	if (DmOpenDatabaseInfo(db, NULL, NULL, NULL, NULL, &res_db) == noErr
	 && res_db != false) {
	    h = DmGet1Resource(HISTORY_RSRC_TYPE, HISTORY_RSRC_ID);
	    if (h == NULL || MemHandleSize(h) != HISTORY_RSRC_SIZE) {
		if (h == NULL) {
		    h = DmNewResource(db, HISTORY_RSRC_TYPE,
			    HISTORY_RSRC_ID, HISTORY_RSRC_SIZE);
		} else {
		    new_h = DmResizeResource(h, HISTORY_RSRC_SIZE);
		    if (new_h == NULL) {
			DmReleaseResource(h);
		    }
		    h = new_h;
		}
		if (h != NULL
		 && (p = MemHandleLock(h)) != NULL) {
		    DmSet(p, 0, HISTORY_RSRC_SIZE, 0);
		    record.entry_size = sizeof(struct history_entry);
		    record.entry_cnt = 0;
		    DmWrite(p, 0, &record, sizeof record);
		    MemHandleUnlock(h);
		}
	    }
	    if (h != NULL) {
		if (MemHandleSize(h) == HISTORY_RSRC_SIZE
		 && (p = MemHandleLock(h)) != NULL) {
		    if (p->entry_size == sizeof(struct history_entry)
		     && p->entry_cnt >= 0
		     && p->entry_cnt <= HISTORY_MAX_COUNT) {
			count = 0;
			for (i = 0; i < p->entry_cnt; ++i) {
			    if (p->entries[i].card_no == entry.card_no
			     && p->entries[i].db_id == entry.db_id) {
				break;
			    }
			    if (p->entries[i].type == entry.type) {
				++count;
				if (count >= HISTORY_EACH_COUNT) {
				    break;
				}
			    }
			}
			if (i == p->entry_cnt) {
			    if (i >= HISTORY_MAX_COUNT) {
				i = HISTORY_MAX_COUNT - 1;
			    } else {
				record = *p;
				++record.entry_cnt;
				DmWrite(p, 0, &record, sizeof record);
			    }
			}
			for ( ; i > 0; --i) {
			    DmWrite(p, sizeof(struct history_record)
				     + sizeof(struct history_entry) * i,
				    &p->entries[i - 1],
				    sizeof(struct history_entry));
			}
			DmWrite(p, sizeof(struct history_record),
				&entry, sizeof entry);
		    }
		    MemHandleUnlock(h);
		}
		DmReleaseResource(h);
	    }
	}
	DmCloseDatabase(db);
    }
}

static void launch_da(UInt card_no, LocalID db_id, ULong stack_size)
{
    DmOpenRef db;
    UInt open_count;
    Boolean res_db;
    VoidHand h;
    void (*da_entry)();
    VoidPtr new_stack;
    long saved_stack;

    db = DmOpenDatabase(card_no, db_id, dmModeReadOnly);
    if (db != NULL) {
	if (DmOpenDatabaseInfo(
		db, NULL, &open_count, NULL, NULL, &res_db) == noErr
	 && open_count == 1
	 && res_db != false
	 && (h = DmGet1Resource(DA_RSRC_TYPE, DA_RSRC_ID)) != NULL) {
	    da_entry = MemHandleLock(h);
	    if (da_entry != NULL) {

		/*
		 * prepare stack
		 */

		new_stack = NULL;
		stack_size &= ~1L;
		if (stack_size >= MIN_STACK_SIZE) {
		    new_stack = MemPtrNew(stack_size);
		}
		saved_stack = __getA7();	/* to shut up gcc -W ... */
		if (new_stack != NULL) {
		    __setA7((long) new_stack + stack_size);
		}

		/*
		 * call the desk accessory
		 */

		InsPtEnable(false);
		(*da_entry)();

		/*
		 * clean up
		 */

		if (new_stack != NULL) {
		    __setA7(saved_stack);
		    MemPtrFree(new_stack);
		}

		MemHandleUnlock(h);
	    }
	    DmReleaseResource(h);
	}
	DmCloseDatabase(db);
    }
}

static void do_launch(ARG(struct globals *g))
{
    UInt attr;
    ULong type;
    ULong creator;

    if (DmDatabaseInfo(g->card_no, g->db_id, g->name, &attr,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, &type, &creator) == noErr
     && (attr & dmHdrAttrResDB) != 0) {
	if (g->prefs.record_history == RECORD_APPS_DAS
	 || g->prefs.record_history == RECORD_ALL_APPS_DAS
	 || ((g->prefs.record_history == RECORD_APPS
	   || g->prefs.record_history == RECORD_ALL_APPS)
	  && type == APP_DB_TYPE)
	 || (g->prefs.record_history == RECORD_DAS
	  && type == DA_DB_TYPE)) {
	    record_history(g->card_no, g->db_id, type, creator, g->name);
	}
	if (type == APP_DB_TYPE) {
	    SysUIAppSwitch(
		    g->card_no, g->db_id, sysAppLaunchCmdNormalLaunch, 0);
	} else if (type == DA_DB_TYPE) {
	    launch_da(g->card_no, g->db_id, g->prefs.stack_size);

	    if (g->saved_form_id != 0) {
		FrmSetActiveForm(g->saved_form);
	    }
	    WinSetActiveWindow(g->saved_active_win);
	    WinSetDrawWindow(g->saved_draw_win);
	}
    }
}

static int find_last_item(ARG_(struct globals *g) int part)
{
    UInt card_no;
    LocalID db_id;
    ULong cur_type, cur_creator, creator;
    int i, ret_code = DO_NOTHING;

    if (part == MODE_LAST_APP) {
	if (SysCurAppDatabase(&card_no, &db_id) != noErr
	 || DmDatabaseInfo(card_no, db_id, NULL, NULL, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, &cur_type, &cur_creator) != noErr) {
	    return ret_code;
	}
	gen_history_list(ARG_(g) MODE_APP_HISTORY, 2, HISTORY_LAST_USED);
    } else {
	gen_history_list(ARG_(g) MODE_DA_HISTORY, 1, HISTORY_LAST_USED);
    }

    for (i = 0; i < g->db_cnt; ++i) {
	g->card_no = g->db_list[i].card_no;
	g->db_id = g->db_list[i].db_id;

	/*
	if (part != MODE_LAST_APP
	 || ((g->card_no != card_no || g->db_id != db_id)
	  && (cur_type != PANEL_DB_TYPE
	   || (DmDatabaseInfo(g->card_no, g->db_id, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, &creator) == noErr
	    && creator != PREFS_CREATOR)))) {
	    ret_code = DO_LAUNCH;
	    break;
	}
	*/

	if (part == MODE_LAST_APP) {
	    if ((g->card_no == card_no && g->db_id == db_id)) {
		continue;
	    }
	    if (DmDatabaseInfo(g->card_no, g->db_id, NULL, NULL, NULL,
		    NULL, NULL, NULL, NULL, NULL, NULL, NULL, &creator) != noErr
	     || (cur_type == PANEL_DB_TYPE
	      && creator == PREFS_CREATOR)
	     || (cur_type == ADVANCED_CALC_TYPE
	      && cur_creator == ADVANCED_CALC_CREATOR
	      && creator == CALC_CREATOR)) {
		continue;
	    }
	}

	ret_code = DO_LAUNCH;
	break;
    }

    free_db_list(ARG(g));

    return ret_code;
}


/*
 * main routines
 */

static void do_launcher(int part)
{
    struct globals ARG_(*g) *gg;
    Err err;
    int ret_code;

    SAVE_A4_REGISTER

    err = FtrGet(CREATOR, FTR_GLOBALS, (DWord *) &gg);
    if (err == noErr) {
	g = gg;
	if (g != NULL && g->creator == CREATOR) {
	    if (part != MODE_LAST_APP) {
		g->new_mode = part;
	    }
	    if (part == MODE_LAST_DA) {
		part = MODE_NONE;
	    }
	}
    } else if (err == ftrErrNoSuchFeature
	    && (part >= MODE_APP && part <= MODE_DA_HISTORY)) {
	g = MemPtrNew(sizeof(struct globals));
	if (g != NULL) {
	    if (FtrSet(CREATOR, FTR_GLOBALS, (DWord) g) == noErr) {
		g->creator = CREATOR;
		get_prefs(&g->prefs);

		GsiSetShiftState(0, 0);

		g->saved_form = FrmGetActiveForm();
		g->saved_form_id = FrmGetActiveFormID();
		g->saved_active_win = WinGetActiveWindow();
		if ((void *) g->saved_form != (void *) g->saved_active_win) {
		    ret_code = DO_NOTHING;
		    goto ret;
		}
		g->saved_draw_win = WinGetDrawWindow();
		g->saved_font = FntSetFont(stdFont);

		g->tap_mode = part;
		do {
		    g->mode = part;
		    g->new_mode = MODE_NONE;
		    ret_code = open_launcher(ARG(g));
		    part = g->new_mode;
		} while (ret_code == SWITCH_MODE);

		if (g->saved_form_id != 0) {
		    FrmSetActiveForm(g->saved_form);
		}
		WinSetActiveWindow(g->saved_active_win);
		WinSetDrawWindow(g->saved_draw_win);
		FntSetFont(g->saved_font);

	    ret:
		FtrUnregister(CREATOR, FTR_GLOBALS);

		if (part == MODE_LAST_DA) {
		    ret_code = find_last_item(ARG_(g) part);
		    part = MODE_NONE;
		}

		if (ret_code == DO_LAUNCH) {
		    do_launch(ARG(g));
		}
	    }
	    MemPtrFree(g);
	}
    }

    if (part == MODE_LAST_APP || part == MODE_LAST_DA) {
	g = MemPtrNew(sizeof(struct globals));
	if (g != NULL) {
	    get_prefs(&g->prefs);

	    ret_code = find_last_item(ARG_(g) part);

	    if (ret_code == DO_LAUNCH && part == MODE_LAST_DA) {
		g->saved_form = FrmGetActiveForm();
		g->saved_form_id = FrmGetActiveFormID();
		g->saved_active_win = WinGetActiveWindow();
		g->saved_draw_win = WinGetDrawWindow();
		if ((void *) g->saved_form != (void *) g->saved_active_win) {
		    ret_code = DO_NOTHING;
		}
	    }

	    if (ret_code == DO_LAUNCH) {
		do_launch(ARG(g));
	    }

	    MemPtrFree(g);
	}
    }

    RESTORE_A4_REGISTER
}

Err MyHandler(PointType *start, PointType *end)
{
    int part;
    Err (*orig_trap)(PointType *start, PointType *end);

    part = find_part(start, end);
    if ((part >= MODE_APP && part <= MODE_LAST_DA)
     && is_device_locked() == false) {
	EvtFlushNextPenStroke();
	do_launcher(part);
    } else if (FtrGet(CREATOR, FTR_ORIG, (DWord *) &orig_trap) == noErr) {
	return (*orig_trap)(start, end);
    }

    return noErr;
}
