/* adal.h */

/*
 * constants
 */

#define noErr 0

#define CREATOR 'ADAL'

#define HISTORY_EACH_COUNT 16
#define HISTORY_MAX_COUNT (HISTORY_EACH_COUNT * 2)

#define HISTORY_CARD_NO 0
#define HISTORY_DB_TYPE 'HIST'
#define HISTORY_DB_NAME "App/DA Launcher History"

#define HISTORY_RSRC_TYPE 'HIST'
#define HISTORY_RSRC_ID 1000
#define HISTORY_RSRC_SIZE (sizeof(struct history_record) \
	    + sizeof(struct history_entry) * HISTORY_MAX_COUNT)

#define OLD_PREF_VERSION 2
#define PREF_VERSION 3

#define OLD_MODE_LAST_APP MODE_APP_HISTORY
#define OLD_MODE_LAST_DA MODE_DA_HISTORY

enum mode {
    MODE_NONE,
    MODE_APP,
    MODE_DA,
    MODE_BOTH,
    MODE_HISTORY,
    MODE_APP_HISTORY,
    MODE_DA_HISTORY,
    MODE_LAST_APP,
    MODE_LAST_DA
};

enum tap_again {
    TAPAGAIN_DO_NOTHING,
    TAPAGAIN_CLEAR_INPUT,
    TAPAGAIN_CLOSE_LAUNCHER,
    TAPAGAIN_OS_LAUNCHER,
    TAPAGAIN_ROTATE_APP_DA,
    TAPAGAIN_ROTATE_DA_APP,
    TAPAGAIN_ROTATE_BOTH,
    TAPAGAIN_LAST_ITEM
};

enum space {
    SPACE_NEXT_ITEM,
    SPACE_INPUT_SPACE
};

enum backspace {
    BACKSPACE_PREVIOUS_ITEM,
    BACKSPACE_CLEAR_CHAR,
    BACKSPACE_CLEAR_INPUT
};

enum page_key {
    PAGEKEY_SCROLL_LINE,
    PAGEKEY_SCROLL_PAGE
};

enum record_history {
    RECORD_OFF,
    RECORD_APPS_DAS,
    RECORD_ALL_APPS_DAS,
    RECORD_APPS,
    RECORD_ALL_APPS,
    RECORD_DAS
};

enum history_order {
    HISTORY_ALPHABETICAL,
    HISTORY_LAST_USED,
    HISTORY_REVERSE_USED
};

enum sort_method {
    SORT_FAST,
    SORT_SLOW,
    SORT_JAPANESE_FIRST,
    SORT_SLOW_JAPANESE_FIRST
};

enum mark_flash_ext {
    MARK_NONE,
    MARK_FLASH,
    MARK_EXT,
    MARK_BOTH
};

#define DEFAULT_BUTTON1 MODE_APP
#define DEFAULT_BUTTON2 MODE_DA
#define DEFAULT_BUTTON3 MODE_HISTORY
#define DEFAULT_BUTTON4 MODE_NONE

#define DEFAULT_TAP_AGAIN TAPAGAIN_CLOSE_LAUNCHER
#define DEFAULT_SPACE SPACE_NEXT_ITEM
#define DEFAULT_BACKSPACE BACKSPACE_CLEAR_INPUT
#define DEFAULT_PAGE_KEY PAGEKEY_SCROLL_PAGE

#define DEFAULT_INPUT_FEEDBACK true
#define DEFAULT_RECORD_HISTORY RECORD_ALL_APPS_DAS

#define DEFAULT_STACK_SIZE 3000
#define STACK_UNIT 1000

#define DEFAULT_DRAG1 MODE_NONE
#define DEFAULT_DRAG2 MODE_NONE
#define DEFAULT_DRAG3 MODE_NONE
#define DEFAULT_DRAG4 MODE_LAST_DA
#define DEFAULT_DRAG5 MODE_LAST_APP

#define DEFAULT_HISTORY_ORDER HISTORY_LAST_USED

#define HISTORY_BASE 6
#define HISTORY_DELTA 4
#define DEFAULT_HISTORY_COUNT 10

#define DEFAULT_APP_NAME_LOOKUP false
#define DEFAULT_DA_NAME_LOOKUP false
#define DEFAULT_SORT_METHOD SORT_FAST
#define DEFAULT_MARK_FLASH_EXT MARK_NONE

#define DEFAULT_RESERVED 255


/*
 * types
 */

struct history_entry {
    UInt card_no;
    UInt db_id_kind;
    LocalID db_id;
    ULong type;
    ULong creator;
    char name[dmDBNameLength];
};

struct history_record {
    Word entry_size;
    Word entry_cnt;

#ifdef __GNUC__
    struct history_entry entries[0];
#else /* __MWERKS__ */
    struct history_entry entries[];
#endif
};

struct adal_prefs {
    Byte button1;
    Byte button2;
    Byte button3;
    Byte button4;

    Byte tap_again;
    Byte space;
    Byte backspace;
    Byte page_key;

    Byte input_feedback;
    Byte record_history;
    Word stack_size;

    Byte drag1;
    Byte drag2;
    Byte drag3;
    Byte drag4;
    Byte drag5;

    Byte reserved;		/* was da_localizer */
    Byte history_order;
    Byte history_count;

    Byte app_name_lookup;
    Byte da_name_lookup;
    Byte sort_method;		/* was multiple_launch */
    Byte mark_flash_ext;	/* was history_check */
};


/*
 * functions (!)
 */

static void get_prefs(struct adal_prefs *prefs)
{
    Byte *p;

    if (PrefGetAppPreferencesV10(
	    CREATOR, PREF_VERSION, prefs, sizeof *prefs) == false) {
	if (PrefGetAppPreferencesV10(CREATOR, OLD_PREF_VERSION,
		prefs, sizeof(struct adal_prefs)) == false) {
	    *((DWord *) &prefs->button1) =
		    (DWord) DEFAULT_BUTTON1 << 24 |
		    (DWord) DEFAULT_BUTTON2 << 16 |
		    (DWord) DEFAULT_BUTTON3 << 8 |
		    (DWord) DEFAULT_BUTTON4;

	    *((DWord *) &prefs->tap_again) =
		    (DWord) DEFAULT_TAP_AGAIN << 24 |
		    (DWord) DEFAULT_SPACE << 16 |
		    (DWord) DEFAULT_BACKSPACE << 8 |
		    (DWord) DEFAULT_PAGE_KEY;

	    *((DWord *) &prefs->input_feedback) =
		    (DWord) DEFAULT_INPUT_FEEDBACK << 24 |
		    (DWord) DEFAULT_RECORD_HISTORY << 16 |
		    (DWord) DEFAULT_STACK_SIZE;

	    *((DWord *) &prefs->drag1) =
		    (DWord) DEFAULT_DRAG1 << 24 |
		    (DWord) DEFAULT_DRAG2 << 16 |
		    (DWord) DEFAULT_DRAG3 << 8 |
		    (DWord) DEFAULT_DRAG4;

	    *((DWord *) &prefs->drag5) =
		    (DWord) DEFAULT_DRAG5 << 24 |
		    (DWord) DEFAULT_RESERVED << 16 |
		    (DWord) DEFAULT_HISTORY_ORDER << 8 |
		    (DWord) DEFAULT_HISTORY_COUNT;

	    *((DWord *) &prefs->app_name_lookup) =
		    (DWord) DEFAULT_APP_NAME_LOOKUP << 24 |
		    (DWord) DEFAULT_DA_NAME_LOOKUP << 16 |
		    (DWord) DEFAULT_SORT_METHOD << 8 |
		    (DWord) DEFAULT_MARK_FLASH_EXT;
	} else {
	    for (p = &prefs->button1; p <= &prefs->drag5; ++p) {
		if (p == &prefs->button4 + 1) {
		    p = &prefs->drag1;
		}
		if (*p == OLD_MODE_LAST_APP) {
		    *p = MODE_LAST_APP;
		} else if (*p == OLD_MODE_LAST_DA) {
		    *p = MODE_LAST_DA;
		}
	    }

	    prefs->reserved = DEFAULT_RESERVED;
	    prefs->sort_method = DEFAULT_SORT_METHOD;
	    prefs->mark_flash_ext = DEFAULT_MARK_FLASH_EXT;
	}
    }
}
