/* adal_panel.c */

#include <Pilot.h>

#include "adal.h"
#include "adal_rcp.h"


/*
 * constants
 */

#define HACKMASTER_FORM_ID 9000

#define BAD_PREF_VALUE 255


/*
 * types
 */

struct pref_item {
    Word popup_id;
    Byte offset;
    Byte default_value;
};

#define OFFSET_OF(field) ((int) &((struct adal_prefs *) 0)->field)

#define SET_ITEM(i, id, field, default_value) \
	*(DWord *) &items[i] = (DWord) (id) << 16 \
		| (DWord) OFFSET_OF(field) << 8 | (DWord) (default_value)

#define SET_ZERO(i) *(DWord *) &items[i] = 0


/*
 * functions
 */

static VoidPtr GetObjectPtr(Word id)
{
    FormPtr form;

    form = FrmGetActiveForm();
    return FrmGetObjectPtr(form, FrmGetObjectIndex(form, id));
}

static void set_popup(Word popup_id, Byte value)
{
    ControlPtr popup;
    ListPtr list;
    CharPtr text;
    Word index;

    popup = GetObjectPtr(popup_id);
    list = GetObjectPtr(popup_id + 1);		/* list_id = popup_id + 1 */

    index = value;
    if (index < LstGetNumberOfItems(list)) {
	text = LstGetSelectionText(list, index);
    } else {
	text = "???";
	index = noListSelection;
    }

    CtlSetLabel(popup, text);
    LstSetSelection(list, index);
}

static void set_popups(struct pref_item *items, struct adal_prefs *prefs)
{
    int i;
    Byte value;

    for (i = 0; items[i].popup_id != 0; ++i) {
	if (items[i].popup_id == HISTORY_COUNT_POPUP_ID) {
	    value = (prefs->history_count - HISTORY_BASE) / HISTORY_DELTA;
	    if (prefs->history_count != HISTORY_BASE + value * HISTORY_DELTA) {
		value = BAD_PREF_VALUE;
	    }
	} else if (items[i].popup_id == ADVANCED_STACK_POPUP_ID) {
	    value = prefs->stack_size / STACK_UNIT;
	    if (value * STACK_UNIT != prefs->stack_size) {
		value = BAD_PREF_VALUE;
	    }
	} else {
	    value = ((Byte *) prefs)[items[i].offset];
	}

	set_popup(items[i].popup_id, value);
    }
}

static void set_defaults(struct pref_item *items)
{
    int i;

    for (i = 0; items[i].popup_id != 0; ++i) {
	set_popup(items[i].popup_id, items[i].default_value);
    }
}

static void set_prefs(struct pref_item *items, struct adal_prefs *prefs)
{
    int i;
    ListPtr list;
    Word value;

    for (i = 0; items[i].popup_id != 0; ++i) {
	list = GetObjectPtr(items[i].popup_id + 1);
	value = LstGetSelection(list);
	if (value != noListSelection) {
	    if (items[i].popup_id == HISTORY_COUNT_POPUP_ID) {
		prefs->history_count = HISTORY_BASE + value * HISTORY_DELTA;
	    } else if (items[i].popup_id == ADVANCED_STACK_POPUP_ID) {
		prefs->stack_size = value * STACK_UNIT;
	    } else {
		((Byte *) prefs)[items[i].offset] = value;
	    }
	}
    }

    PrefSetAppPreferencesV10(CREATOR, PREF_VERSION, prefs, sizeof *prefs);
}

static void clear_history(void)
{
    DmSearchStateType state;
    UInt card_no;
    LocalID db_id;

    if (DmGetNextDatabaseByTypeCreator(true, &state,
	    HISTORY_DB_TYPE, CREATOR, true, &card_no, &db_id) == noErr) {
	if (FrmAlert(CLEAR_HISTORY_ALERT_ID) == CLEAR_HISTORY_OK_INDEX) {
	    DmDeleteDatabase(card_no, db_id);
	}
    }
}

static void set_list_choices(Word id)
{
    ListPtr list;

    list = GetObjectPtr(GENERAL_BUTTON1_LIST_ID);
    LstSetListChoices(GetObjectPtr(id), list->itemsText, list->numItems);
}

static void do_dialog(Word form_id)
{
    FormPtr form, saved_form;
    struct pref_item items[10];
    struct adal_prefs prefs;
    EventType event;

    saved_form = FrmGetActiveForm();

    form = FrmInitForm(form_id);
    FrmSetActiveForm(form);

    switch (form_id) {
    case GENERAL_DIALOG_ID:
	set_list_choices(GENERAL_BUTTON2_LIST_ID);
	set_list_choices(GENERAL_BUTTON3_LIST_ID);
	set_list_choices(GENERAL_BUTTON4_LIST_ID);
	set_list_choices(GENERAL_DRAG1_LIST_ID);
	set_list_choices(GENERAL_DRAG2_LIST_ID);
	set_list_choices(GENERAL_DRAG3_LIST_ID);
	set_list_choices(GENERAL_DRAG4_LIST_ID);
	set_list_choices(GENERAL_DRAG5_LIST_ID);

	SET_ITEM(0, GENERAL_BUTTON1_POPUP_ID, button1, DEFAULT_BUTTON1);
	SET_ITEM(1, GENERAL_BUTTON2_POPUP_ID, button2, DEFAULT_BUTTON2);
	SET_ITEM(2, GENERAL_BUTTON3_POPUP_ID, button3, DEFAULT_BUTTON3);
	SET_ITEM(3, GENERAL_BUTTON4_POPUP_ID, button4, DEFAULT_BUTTON4);
	SET_ITEM(4, GENERAL_DRAG1_POPUP_ID, drag1, DEFAULT_DRAG1);
	SET_ITEM(5, GENERAL_DRAG2_POPUP_ID, drag2, DEFAULT_DRAG2);
	SET_ITEM(6, GENERAL_DRAG3_POPUP_ID, drag3, DEFAULT_DRAG3);
	SET_ITEM(7, GENERAL_DRAG4_POPUP_ID, drag4, DEFAULT_DRAG4);
	SET_ITEM(8, GENERAL_DRAG5_POPUP_ID, drag5, DEFAULT_DRAG5);
	SET_ZERO(9);
	break;

    case INPUT_DIALOG_ID:
	SET_ITEM(0, INPUT_TAPAGAIN_POPUP_ID, tap_again, DEFAULT_TAP_AGAIN);
	SET_ITEM(1, INPUT_SPACE_POPUP_ID, space, DEFAULT_SPACE);
	SET_ITEM(2, INPUT_BACKSPACE_POPUP_ID, backspace, DEFAULT_BACKSPACE);
	SET_ITEM(3, INPUT_PAGEKEY_POPUP_ID, page_key, DEFAULT_PAGE_KEY);
	SET_ITEM(4, INPUT_FEEDBACK_POPUP_ID, input_feedback,
		DEFAULT_INPUT_FEEDBACK);
	SET_ZERO(5);
	break;

    case HISTORY_DIALOG_ID:
	SET_ITEM(0, HISTORY_RECORD_POPUP_ID, record_history,
		DEFAULT_RECORD_HISTORY);
	SET_ITEM(1, HISTORY_ORDER_POPUP_ID, history_order,
		DEFAULT_HISTORY_ORDER);
	SET_ITEM(2, HISTORY_COUNT_POPUP_ID, history_count,
		(DEFAULT_HISTORY_COUNT - HISTORY_BASE) / HISTORY_DELTA);
	SET_ZERO(3);
	break;

    case ADVANCED_DIALOG_ID:
	SET_ITEM(0, ADVANCED_APP_NAME_POPUP_ID, app_name_lookup,
		DEFAULT_APP_NAME_LOOKUP);
	SET_ITEM(1, ADVANCED_DA_NAME_POPUP_ID, da_name_lookup,
		DEFAULT_DA_NAME_LOOKUP);
	SET_ITEM(2, ADVANCED_SORT_POPUP_ID, sort_method, DEFAULT_SORT_METHOD);
	SET_ITEM(3, ADVANCED_MARK_POPUP_ID, mark_flash_ext,
		DEFAULT_MARK_FLASH_EXT);
	SET_ITEM(4, ADVANCED_STACK_POPUP_ID, stack_size,
		DEFAULT_STACK_SIZE / STACK_UNIT);
	SET_ZERO(5);
	break;
    }

    get_prefs(&prefs);
    set_popups(items, &prefs);

    FrmDrawForm(form);

    for (;;) {
	EvtGetEvent(&event, evtWaitForever);

	if (SysHandleEvent(&event) == false) {
	    if (event.eType == appStopEvent) {
		EvtAddEventToQueue(&event);
		goto ret;
	    }

	    if (FrmHandleEvent(form, &event) == false) {
		if (event.eType == ctlSelectEvent) {
		    switch (event.data.ctlSelect.controlID) {
		    case GENERAL_OK_ID:
		    case INPUT_OK_ID:
		    case HISTORY_OK_ID:
		    case ADVANCED_OK_ID:
			set_prefs(items, &prefs);

			/* fall through */

		    case GENERAL_CANCEL_ID:
		    case INPUT_CANCEL_ID:
		    case HISTORY_CANCEL_ID:
		    case ADVANCED_CANCEL_ID:
			goto ret;

		    case GENERAL_DEFAULT_ID:
		    case INPUT_DEFAULT_ID:
		    case HISTORY_DEFAULT_ID:
		    case ADVANCED_DEFAULT_ID:
			set_defaults(items);
			break;

		    case HISTORY_CLEAR_ID:
			clear_history();
			break;
		    }
		}
	    }
	}
    }

ret:
    FrmEraseForm(form);
    FrmDeleteForm(form);
    FrmSetActiveForm(saved_form);
}

Boolean MyHandler(EventPtr event)
{
    Boolean handled = false;
    FormPtr form;

    switch (event->eType) {
    case frmOpenEvent:
	form = FrmGetActiveForm();
	FrmDrawForm(form);
	handled = true;
	break;

    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case PANEL_GENERAL_ID:
	    do_dialog(GENERAL_DIALOG_ID);
	    handled = true;
	    break;

	case PANEL_INPUT_ID:
	    do_dialog(INPUT_DIALOG_ID);
	    handled = true;
	    break;

	case PANEL_HISTORY_ID:
	    do_dialog(HISTORY_DIALOG_ID);
	    handled = true;
	    break;

	case PANEL_ADVANCED_ID:
	    do_dialog(ADVANCED_DIALOG_ID);
	    handled = true;
	    break;

	case PANEL_OK_ID:
	    FrmGotoForm(HACKMASTER_FORM_ID);
	    handled = true;
	    break;
	}
	break;

    default:		/* to shut up gcc -W ... */
	break;
    }

    return handled;
}
